/*
Author:       Pascal Peter
Copyright:    (C) 2020 Pascal Peter
License:      GNU General Public License version 3
Website:      http://pascal.peter.free.fr
Email:        pascal.peter at free.fr
*/


const SHAPES_DEF = [
    [], 
    [[0,1],[1,0],[1,1]], 
    [[1,0]], 
    [[0,1]], 
    [[1,-1],[1,0]], 
    [[1,0],[1,1]], 
    [[0,1],[1,1]], 
    [[0,1],[1,0]], 
    [[1,0],[2,0]], 
    [[0,1],[0,2]]];//y,x

const BORDERS_DEF = 
    [['0123'], 
    ['03','01','23','12'], 
    ['013','123'], 
    ['023','012'], 
    ['013','023','12'], 
    ['013','23','012'], 
    ['023','01','123'], 
    ['03','012','123'], 
    ['013','13','123'], 
    ['023','02','012']];//0:top, 1:right, 2:bottom, 3:left

let SIZE;




function init() {

    let grid;
    do {
        grid = getGrid(0, [], 0, []);
    }
    while (grid == null);

    let gridContent = '', 
        solutionContent = '';
    for (let r = 0; r < SIZE; r++) {
        gridContent += '<tr>';
        solutionContent += '<tr>';
        for (let c = 0; c < SIZE; c++) {
            gridContent += '<td id="r' + r + 'c' + c + '"><span><\/span>';
            gridContent += '<input class="grid-input" type="text" onfocus="javascript:this.select();"\/><\/td>';
            solutionContent += '<td id="sol_r' + r + 'c' + c + '">' + grid[r][c] + '<\/td>';
        }
        gridContent += '<\/tr>';
        solutionContent += '<\/tr>';
    }
    gridContent = '<table class="grid-table">' + gridContent + '<\/table>';
    solutionContent = '<table class="grid-table" id="grid-solution" style="display:none;">' + solutionContent + '<\/table>';
    document.getElementById('content').innerHTML = gridContent;
    document.getElementById('solution').innerHTML = solutionContent;

    let shapes = getShapes();
    for (let shape in shapes)
        fillShape(grid, shape, shapes[shape]);
}

function getShapes() {
    let cells = [];
    for (let i = 0; i < SIZE * SIZE; i++)
        cells[i] = true;

    let shapes = {};
    for (let r = 0; r < SIZE; r++) {
        for (let c = 0; c < SIZE; c++) {
            if (cells[r * SIZE + c]) {
                let sid, 
                    coords, 
                    badShape, 
                    iter = 0;
                do {
                    iter++;
                    sid = iter < 100 ? getRand(8) + 1 : 0;
                    coords = [[0, 0]].concat(SHAPES_DEF[sid]);
                    badShape = false;

                    for (let co in coords) {
                        let x = c + coords[co][1], 
                            y = r + coords[co][0];
                        if (x < 0 || y < 0 || x >= SIZE || y >= SIZE || !cells[y * SIZE + x]) {
                            badShape = true;
                            break;
                        }
                    }
                }
                while(badShape);

                for(let co in coords)
                    cells[(r+coords[co][0])*SIZE+(c+coords[co][1])] = false;
                shapes[r*SIZE+c] = sid;
            }
        }
    }
    return shapes;
}

//Fill each shape with the appropriate operations/result and borders
function fillShape(grid, org, type) {
    let coords = [[0, 0]].concat(SHAPES_DEF[type]), 
        borders = BORDERS_DEF[type], 
        orgX = org % SIZE, 
        orgY = Math.floor(org / SIZE);

    let values = [];
    for (let co in coords) {
        let y = orgY + coords[co][0], 
            x = orgX + coords[co][1], 
            bor = 'b' + borders[co].split('').join(' b');
        document.getElementById('r' + y + 'c' + x).className = bor;
        document.getElementById('sol_r' + y + 'c' + x).className = 'sol '+ bor;
        values.push(grid[y][x]);
    }
    document.getElementById('r' + orgY + 'c' + orgX).children[0].innerHTML = getRandOp(values);
}

//Returns a random grid with each number being unique on every rows and columns
function getGrid(r, rows, c, cols) {
    if (r == SIZE && c == 0)
        return rows;
    else {
        if (rows[r] == null)
            rows[r] = [];
        if ( cols[c] == null)
            cols[c] = [];
        let n = getUniqueRand(rows[r], cols[c]);
        if (n == -1)
            return null;
        else {
            rows[r][c] = n;
            cols[c][r] = n;
            if (c == SIZE - 1)
                r++;
            return getGrid(r, rows, (c + 1) % SIZE, cols);
        }
    }
}

//Returns a random operation result and sign based on array values
function getRandOp(values) {
    if (values.length == 1)
        return values[0];
    else if (values.length == 2)
        values = [Math.max(values[0], values[1]), Math.min(values[0], values[1])];

    let isDecimalDiv = values[0] % values[1] != 0, 
        op;
    do {
        op = getRand(3);
    }
    while (((op == 1 || op == 3) && values.length != 2) || (op == 3 && isDecimalDiv));

    let ch = '+-*/'.charAt(op), 
        ch2 = '+-×÷'.charAt(op);
    return eval(values.join(ch)) + ch2;
}

//Returns a unique (according to column and row data) random number
function getUniqueRand(col, row) {
    let iter = 0, 
        iterMax = Math.pow(SIZE, 2), 
        ur;
    do {
        if (iter > iterMax)
            return -1;
        ur = getRand(SIZE - 1) + 1;
        iter++;
    }
    while(inArray(col, ur) || inArray(row, ur));
    return ur;
}

function inArray(array, value) {
    return (array.indexOf(value) > -1);
}

function getRand(max) {
    return Math.round(Math.random() * max);
}














function toggleSolution() {
    let element = document.getElementById('grid-solution');
    if (element.style.display == 'none')
        element.style.display = 'table';
    else
        element.style.display = 'none';
}

function reset() {
    SIZE = parseInt(document.getElementById('size').value);
    init();
}

window.onload = reset();
document.getElementById('reset').addEventListener('click', reset, false);
document.getElementById('resolve').addEventListener('click', toggleSolution, false);
document.getElementById('size').addEventListener('change', reset, false);





